"""
Enum of the validator.

Author: Fabian A. Preiss
"""
from enum import Enum


class AngleSelectParams(Enum):
    IMG = "IMG"
    REF_IMG = "REF_IMG"
    ANGLE_A = "ANGLE_A"
    ANGLE_B = "ANGLE_B"
    UPSAMPLING = "UPSAMPLING"
    ROT_A_IMG = "ROT_A_IMG"
    ROT_B_IMG = "ROT_B_IMG"
    TRANSLATION_A = "TRANSLATION_A"
    TRANSLATION_B = "TRANSLATION_B"
    ROT_TR_A_IMG = "ROT_TR_A_IMG"
    ROT_TR_B_IMG = "ROT_TR_B_IMG"
    SELECTOR = "SELECTOR"
    RECOVERED_ROTATION = "RECOVERED_ROTATION"
    RECOVERED_ROT_IMG = "RECOVERED_ROT_SCALE_IMG"
    RECOVERED_TRANSLATION = "RECOVERED_TRANSLATION"
    RECOVERED_ROT_TR_IMG = "RECOVERED_ROT_TR_IMG"
