"""
Enum of the logpolar parameters.

Author: Fabian A. Preiss
"""
from enum import Enum


class LogPolParams(Enum):
    GAUSS_DIFF = "GAUSS_DIFF"
    WINDOW_WEIGHT = "WINDOW_WEIGHT"
    WINDOW_TYPE = "WINDOW_TYPE"
    WINDOW_RADIUS_EXP = "WINDOW_RADIUS_EXP"
    BOUNDS = "BOUNDS"
    REF_IMG = "REF_IMG"
    MOD_IMG = "MOD_IMG"
    GAUSS_DIFF_REF_IMG = "GAUSS_DIFF_REF_IMG"
    GAUSS_DIFF_MOD_IMG = "GAUSS_DIFF_MOD_IMG"
    FOURIER_REF_IMG = "FOURIER_REF_IMG"
    FOURIER_MOD_IMG = "FOURIER_MOD_IMG"
    WARPED_FOURIER_REF_IMG = "WARPED_FOURIER_REF_IMG"
    WARPED_FOURIER_MOD_IMG = "WARPED_FOURIER_MOD_IMG"
    UPSAMPLING = "UPSAMPLING"
    RECOVERED_ROTATION_SCALE_PHASE = "RECOVERED_ROTATION_SCALE_PHASE"
    RECOVERED_ROTATION = "RECOVERED_ROTATION"
    RECOVERED_SCALE = "RECOVERED_SCALE"
    RECOVERED_TRANSLATION = "RECOVERED_TRANSLATION"
    RECOVERED_ROT_SCALE_IMG = "RECOVERED_ROT_SCALE_IMG"
    RECOVERED_ROT_SCALE_TR_IMG = "RECOVERED_ROT_SCALE_TR_IMG"
