"""
Enum of the radon parameters.

Author: Fabian A. Preiss
"""
from enum import Enum


class RadonParams(Enum):
    REF_IMG = "REF_IMG"
    MOD_IMG = "MOD_IMG"
    ANGULAR_PRECISION = "ANGULAR_PRECISION"
    EXPONENTIAL_FILTER_SIGNAL_NOISE = "EXPONENTIAL_FILTER_SIGNAL_NOISE"
    THETA = "THETA"
    UPSAMPLING = "UPSAMPLING"
    REF_ROTATION = "REF_ROTATION"
    MOD_ROTATION = "MOD_ROTATION"
    ROTATION_CANDIDATE = "ROTATION_CANDIDATE"
    ANGLE_SELECT = "ANGLE_SELECT"
    RECOVERED_ROTATION = "RECOVERED_ROTATION"
    RECOVERED_ROT_IMG = "RECOVERED_ROT_SCALE_IMG"
    RECOVERED_TRANSLATION = "RECOVERED_TRANSLATION"
    RECOVERED_ROT_TR_IMG = "RECOVERED_ROT_TR_IMG"
